/* ----------------------------------------------------------------------------
 * main()
 * ----------------------------------------------------------------------------
 * Copyright (c) 2012 Ben Blazak <benblazak.dev@gmail.com>
 * Released under The MIT License (MIT) (see "license.md")
 * Project located at <https://github.com/benblazak/ergodox-firmware>
 * ------------------------------------------------------------------------- */


#include <stdbool.h>
#include <stdint.h>

// --------------------------------------------------------------------
// hardware
// --------------------------------------------------------------------

#define KBD_DEBUG // comment out to disable the debug interface completely
#include "./keyboard/controller.c"
#include "./keyboard/keyboard.h"

// --------------------------------------------------------------------
// types and forward declarations
// --------------------------------------------------------------------

typedef uint8_t u8;
typedef uint16_t u16;

typedef u8 	keycode;
typedef u16	media_keycode;
typedef u8 	layer;

typedef void (*keyfunc)(void);

#include "./main.h"

// ----------------------------------------------------------------------------
// layout data
// ----------------------------------------------------------------------------

#include "./keyboard/layout.c"
// defines:
// #define KB_Layers #{Layers.size}
// static const keycode	PROGMEM _kb_layout[KB_LAYERS][KB_ROWS][KB_COLUMNS];
// static const keyfunc	PROGMEM _kb_layout_press[KB_LAYERS][KB_ROWS][KB_COLUMNS];
// static const keyfunc	PROGMEM _kb_layout_release[KB_LAYERS][KB_ROWS][KB_COLUMNS];

// ----------------------------------------------------------------------------
// globals
// ----------------------------------------------------------------------------

static bool	_kb_is_pressed[KB_ROWS][KB_COLUMNS];
static bool	(*kb_is_pressed)[KB_ROWS][KB_COLUMNS] = &_kb_is_pressed;

static bool	_kb_was_pressed[KB_ROWS][KB_COLUMNS];
static bool	(*kb_was_pressed)[KB_ROWS][KB_COLUMNS] = &_kb_was_pressed;

static layer	layers_pressed[KB_ROWS][KB_COLUMNS];

static u8     	current_row;
static u8     	current_col;
static layer  	current_layer;
static keycode	current_keycode;
static bool   	current_is_pressed;

static bool 	layers_active[KB_LAYERS];
static layer	layers_top = 0;

static bool	sticky_on;
static bool	sticky_done;
static bool	layer_sticky[KB_LAYERS];
static u8  	modifier_sticky;

// ----------------------------------------------------------------------------

int main() {
  kb_init();
  usb_init();
  while (!usb_configured());

  // initialize
  init_layers();
  init_sticky();

  // never return
  main_key_loop();
  return 0;
}

// --------------------------------------------------------------------------------------

void main_key_loop() {
  for (;;) {
    // swap `kb_is_pressed` and `kb_was_pressed`, then update
    bool (*temp)[KB_ROWS][KB_COLUMNS]	= kb_was_pressed;
    kb_was_pressed                   	= kb_is_pressed;
    kb_is_pressed                    	= temp;

    kb_update_matrix(*kb_is_pressed);

    // - execute key functions when their key changes state
    // - keep track of which layers the keys were on when they were pressed
    //   (so they can be released using the function from that layer)
    for (u8 row=0; row<KB_ROWS; row++) {
      for (u8 col=0; col<KB_COLUMNS; col++) {
        current_is_pressed	= (*kb_is_pressed)[row][col];
        bool was_pressed  	= (*kb_was_pressed)[row][col];

        if (current_is_pressed != was_pressed) {
          if (current_is_pressed) {
            current_layer           	= layers_top;
            layers_pressed[row][col]	= current_layer;
          } else {
            current_layer	= layers_pressed[row][col];
          }

          // set remaining vars, and "execute" key
          current_row    	= row;
          current_col    	= col;
          current_keycode	= kb_keycode(current_layer, current_row, current_col);
          exec_key();
        }
      }
    }

    // send the USB report (even if nothing's changed)
    usb_keyboard_send();
    usb_extra_consumer_send();

    // unset sticky keys if necessary
    if (sticky_on && sticky_done) {
      sticky_disable();
      usb_keyboard_send();
    }

    // debounce in ms; see keyswitch spec for necessary value
    _delay_ms(5);
  }
}

// ----------------------------------------------------------------------------
// layer functions
// ----------------------------------------------------------------------------

void init_layers() {
  for (layer l=0; l < KB_LAYERS; l++) {
    layers_active[l] = false;
  }
  layers_active[0] = true;
}


// find highest active layer
layer highest_active_layer(layer offset) {
  if (offset < layers_top) {
    for (layer l = layers_top - offset; l > 0 && l < KB_LAYERS; l--) {
      if (layers_active[l]) { return l; }
    }
  }

  // the base layer is always active
  return 0;
}

// enable a layer
void layer_enable(layer l) {
  if (l >= KB_LAYERS) { return; }

  layers_active[l] = true;

  if (l > layers_top) {
    layers_top = l;
  }
}

// disable a layer
void layer_disable(layer l) {
  // base layer stays always on
  if (l >= KB_LAYERS || l == 0) { return; }

  layers_active[l] = false;

  if (l == layers_top) {
    layers_top = highest_active_layer(1);
  }
}

// disable the highest active layer
void layer_disable_top() {
  layer_disable(layers_top);
}

// return layer offset elements below the top
layer layer_peek(layer offset) {
  return highest_active_layer(offset);
}

bool is_layer_enable(keyfunc f) {
  if (f == &kbfun_layer_enable || f == &kbfun_layer_sticky) {
    return true;
  }
  return false;
}

bool is_layer_disable(keyfunc f) {
  if (f == &kbfun_layer_disable || f == &kbfun_layer_sticky) {
    return true;
  }
  return false;
}

void layer_enable_upto(layer max_layer) {
  // FIXME clean this up

  // pressing a key implicitly activates all lower layers as well
  for (layer l=0; l <= KB_LAYERS; l++) {
    void (*key_function)(void) = kb_keyfunc_press(l, current_row, current_col);

    if (is_layer_enable(key_function)) {
      layer enable_layer = (layer) kb_keycode(l, current_row, current_col);
      if (enable_layer <= max_layer) {
        layer_enable(enable_layer);
      }
    }
  }
}

// ----------------------------------------------------------------------------
// sticky functions
// ----------------------------------------------------------------------------

void init_sticky() {
  for (layer l=1; l < KB_LAYERS; l++) {
    layer_sticky[l] = false;
  }
  for (keycode mod=0; mod < MODIFIERS; mod++) {
    modifier_sticky = 0;
  }
  sticky_on  	= false;
  sticky_done	= true;
}

void sticky_disable() {
  for (layer l=1; l < KB_LAYERS; l++) {
    if (layer_sticky[l]) {
      layer_disable(l);
      layer_sticky[l] = false;
    }
  }

  if (modifier_sticky) {
    keyboard_modifier_keys &= ~modifier_sticky;
    modifier_sticky = 0;
  }

  sticky_on  	= false;
  sticky_done	= false;
}

// ----------------------------------------------------------------------------
// layout info
// ----------------------------------------------------------------------------

keycode kb_keycode        	(layer l, u8 row, u8 col)	{ return (keycode) pgm_read_byte(&(_kb_layout[l][row][col])); }
keyfunc kb_keyfunc_press  	(layer l, u8 row, u8 col)	{ return (keyfunc) pgm_read_word(&(_kb_layout_press[l][row][col])); }
keyfunc kb_keyfunc_release	(layer l, u8 row, u8 col)	{ return (keyfunc) pgm_read_word(&(_kb_layout_release[l][row][col])); }

// ----------------------------------------------------------------------------
// keyfunc primitives
// ----------------------------------------------------------------------------

// basic keypresses
void _kbfun_normal_press_release(bool press, keycode key) {
  if (key == 0) { return; } // noop

  if (press) {
    for (u8 i=0; i < sizeof(keyboard_keys); i++) {
      if (keyboard_keys[i] == 0) {
        keyboard_keys[i] = key;
        return;
      }
    }
  } else {
    for (u8 i=0; i < sizeof(keyboard_keys); i++) {
      if (keyboard_keys[i] == key) {
        keyboard_keys[i] = 0;
        return;
      }
    }
  }
}

void _kbfun_mediakey_press_release(bool press, keycode key) {
  media_keycode media_key = _media_code_lookup_table[key];
  if (press) {
    consumer_key = media_key;
  } else {
    // only one media key can be pressed at a time, so only clear most recent one
    if (media_key == consumer_key) {
      consumer_key = 0;
    }
  }
}

void _kbfun_modifier_press_release(bool press, keycode key) {
  if (press) {
    keyboard_modifier_keys |= (1<<key);
  } else {
    keyboard_modifier_keys &= ~(1<<key);
  }
}

bool _kbfun_normal_is_pressed(keycode key) {
  for (u8 i=0; i < sizeof(keyboard_keys); i++) {
    if (keyboard_keys[i] == key) {
      return true;
    }
  }

  return false;
}

bool _kbfun_mediakey_is_pressed(keycode key) {
  return (consumer_key != 0);
}

bool _kbfun_modifier_is_pressed(keycode key) {
  return (keyboard_modifier_keys & (1<<key));
}

// ----------------------------------------------------------------------------
// basic keyfuncs
// ----------------------------------------------------------------------------

// execute the keypress or keyrelease function (if it exists) of the key at the current possition
void exec_key() {
  void (*key_function)(void) =
    ( (current_is_pressed)
      ? kb_keyfunc_press(current_layer, current_row, current_col)
      : kb_keyfunc_release(current_layer, current_row, current_col) );

  if (key_function) { (*key_function)(); }
}

// normal key
void kbfun_normal_press_release() {
  sticky_done = true;
  _kbfun_normal_press_release(current_is_pressed, current_keycode);
}

// media key
void kbfun_mediakey_press_release() {
  sticky_done = true;
  _kbfun_mediakey_press_release(current_is_pressed, current_keycode);
}

// modifier
void kbfun_modifier_press_release() {
  _kbfun_modifier_press_release(current_is_pressed, current_keycode);
}

// enable layer
void kbfun_layer_enable() {
  sticky_done = true;

  layer l = (layer) current_keycode;
  layer_enable_upto(l);
}

// disable layer
void kbfun_layer_disable() {
  sticky_done = true;

  // letting go off a key releases *all* layers on that key
  for (layer l=0; l <= KB_LAYERS; l++) {
    // FIXME this is broken with sticky
    void (*key_function)(void) = kb_keyfunc_release(l, current_row, current_col);

    if (is_layer_disable(key_function)) {
      layer disable_layer = (layer) kb_keycode(l, current_row, current_col);
      layer_disable(disable_layer);
    }
  }
}

// sticky layer key
void kbfun_layer_sticky() {
  layer l = (layer) current_keycode;

  if (current_is_pressed) {
    layer_enable(l);
    sticky_done = false;
  } else {
    if (sticky_done) {
      layer_disable(l);
    } else {
      layer_sticky[l]	= true;
      sticky_on      	= true;
      sticky_done    	= false;
    }
  }
}

// sticky modifier key
void kbfun_modifier_sticky() {
  // TODO handle: sticky, then same modifier
  // TODO split sticky_on/done layer vs modifier so we can fine-tune it more?
  keycode mod = current_keycode;

  if (current_is_pressed) {
    kbfun_modifier_press_release();
    sticky_done = false;
  } else {
    if (sticky_done) {
      kbfun_modifier_press_release();
    } else {
      modifier_sticky	|= 1<<mod;
      sticky_on      	= true;
      sticky_done    	= false;
    }
  }
}

// ----------------------------------------------------------------------------
// combo keyfuncs
// ----------------------------------------------------------------------------

void _kbfun_combo_press_release(keycode combo_key) {
  _kbfun_modifier_press_release(current_is_pressed, combo_key);
  kbfun_normal_press_release();
}

void kbfun_shift_press_release()  	{ _kbfun_combo_press_release(MOD_KEY_LeftShift); }  	// +shift
void kbfun_control_press_release()	{ _kbfun_combo_press_release(MOD_KEY_LeftControl); }	// +control
void kbfun_alt_press_release()    	{ _kbfun_combo_press_release(MOD_KEY_LeftAlt); }    	// +alt
void kbfun_win_press_release()    	{ _kbfun_combo_press_release(MOD_KEY_LeftGUI); }    	// +win

// capslock
void kbfun_capslock_press_release() {
  static u8  	keys_pressed;
  static bool	lshift_pressed;
  static bool	rshift_pressed;

  keycode key = current_keycode;

  if (!current_is_pressed) { keys_pressed--; }

  // take care of the key that was actually pressed
  _kbfun_modifier_press_release(current_is_pressed, key);

  // take care of capslock (only on the press of the 2nd key)
  if (keys_pressed == 1 && current_is_pressed) {
    // save the state of left and right shift
    lshift_pressed = _kbfun_modifier_is_pressed(MOD_KEY_LeftShift);
    rshift_pressed = _kbfun_modifier_is_pressed(MOD_KEY_RightShift);
    // disable both
    _kbfun_modifier_press_release(false, MOD_KEY_LeftShift);
    _kbfun_modifier_press_release(false, MOD_KEY_RightShift);

    // press capslock, then release it
    _kbfun_normal_press_release(true, 	KEY_CapsLock);	usb_keyboard_send();
    _kbfun_normal_press_release(false,	KEY_CapsLock);	usb_keyboard_send();

    // restore the state of left and right shift
    if (lshift_pressed)	{ _kbfun_modifier_press_release(true, MOD_KEY_LeftShift); 	}
    if (rshift_pressed)	{ _kbfun_modifier_press_release(true, MOD_KEY_RightShift);	}
  }

  if (current_is_pressed) { keys_pressed++; }
}
